# # #
# /5/algo.rb
# (c) 2009 Jan Lelis <mail@janlelis.de>. MIT License.
# See: https://rbjl.janlelis.com/5-playing-with-dijkstra
# # #

require 'set'

# now it gets interesting: this module holds the algorithms
module Algo
  # each Algo has its own namespace
  module Dijkstra
   NeedsGraph = true
   class << self # everything are module methods - dont write self. prefix everytime..

    def info # is displayed as console info
      'Shortest path length: ' << @path.length.to_s
    end

    # init dijkstra, if an heuristic is given, how to decide [by min] which sould be the next node, than use it!
    def init(nodes,edges,&heuristic)
      @a, @edges, @heuristic = nodes, edges, heuristic||proc{|sn| sn[1].d} # std heuristic is dijkstra [next node is the one with min. distance]
      @m = {} # marked nodes
      @s = {} # sidenodes
    end

    # calculate the solution!
    def calc()
      # get shortest distance tupel :)
      current = DTupel.new(0,0,0) # 0 is startnode
      until current.n == 80 or @a.empty?
        @a.delete current.n # delete node from @a
        @s.delete current   # .. and from @s
        @m[current.n] = current # add tupel to @m

        new_tupels = create_tupels current # get new sidenode-tupels
        new_tupels.each do |sn|
          @s[sn.n] = sn  if !@s[sn.n] or @s[sn.n].d > sn.d # add sidenode if it does not exist or choose the one with smallest distance
        end
        # check if there are sidenodes
        return '*No solution path is possible in this labyrinth!*' if @s.empty?
        # get minimum for next iteration
        next_node = @s.min_by(&@heuristic) # classical dijkstra: {|sn| sn[1].d }
        current = next_node[1]
        @s.delete next_node[0]
      end

      # get final path
      ret = []
      until current.n == 0
        ret << get_direction(current.n, current.p)
        current = @m[current.p]
      end
      @path = ret.reverse
    end

    # create new sidenode tupels
    def create_tupels(tupel)
      # get open edges
      # comment out next line for JRuby 1.2.0 support (see readme!)
      open_edges = @edges.select {|cur,other| tupel.n==cur and @a.include? other} # connection between current and a still available ones - this is the normal way, working in 1.9
      # remove comment next line for JRuby 1.2.0 support (see readme!)
      # open_edges = []; @edges.each{|cur,other| open_edges << [cur,other] if tupel.n==cur&&@a.include?(other)}
      open_edges.collect {|e| DTupel.new(e[1],tupel.d+1,tupel.n) }
    end

    # calculate direction by node identifier
    def get_direction(n1,n2)
      case
        when n2 - n1 == 1 then 'l' # left
        when n1 - n2 == 1 then 'r' # right
        when n2 - n1 == 9 then 'u' # up
        when n1 - n2 == 9 then 'd' # down
      end
    end

   end

   # One tupel takes @n,@d,@p -> current node, current (shortest) distance, prior node
   class DTupel
     attr_reader :n,:d,:p
     attr_writer    :d,:p

     def initialize(n,d=1.0/0,p=nil) # d = Infinity xD
       @n,@d,@p = n,d,p #node, distance, prior node
     end
   end

  end


 module Backtracking
    NeedsGraph = false
    class << self

      def info
        "This algorithm does not calculate the shortest path!\nIt's calculating *a* path to the finish!\nPath length: #{@path.length}"
      end

      def init(a)
        @a = a;
        @coords = [[0,0]]
        @path   = []
      end

      def calc()
        catch :found do
          search 0,0,@coords # backtrack for solutions (-->coords) and stop! if one is found
          return '*No solution path is possible in this labyrinth!*' if @coords.last != [8,8]
        end

        @coords.each_cons(2) {|old,new| @path << get_direction(old,new)}
        @path
      end

      def consistent(x,y,coords)
        (0..8).member? x and (0..8).member? y and @a[x][y] == 0 and not coords.member?([x,y])
      end

      def search(x,y,coords)
        if x==8&&y==8 # reached finish! abort..
          @coords = coords
          throw :found
        end
        (x+=1;coords << [x,y];search x,y,coords.clone) if consistent(x+1,y,coords)
        (y+=1;coords << [x,y];search x,y,coords.clone) if consistent(x,y+1,coords)
        (x-=1;coords << [x,y];search x,y,coords.clone) if consistent(x-1,y,coords)
        (y-=1;coords << [x,y];search x,y,coords.clone) if consistent(x,y-1,coords)
     end

     def get_direction(old,new) # ...between two points
       if old[0] == new[0]
         if old[1] == new[1]-1 then  'r' # right
         else                        'l' # left
         end
       elsif old[0] == new[0]-1 then 'd' # down
       else                          'u' # up
       end
     end

   end
 end
end
